// OpenCP Carriage Return Killer
// $Header: /code/killcr/killcr.cc,v 1.18 2007/06/21 11:07:22 doj Exp $
//
// copyright (c) 1998-2007 Dirk Jagdmann <doj@cubic.org>
//
// This program kills all <CR> in a text file

/***************************************************************************
 *                                                                         *
 *                        CONFIGURE SECTION                                *
 *                                                                         *
 ***************************************************************************/

/// all suffixes which are parsed (not case sensitive)
const char *suffix[] = {
  ".asm",
  ".bat",
  ".bib",
  ".c",
  ".cc",
  ".cpp",
  ".css",
  ".csv",
  ".cxx",
  ".diz",
  ".dtd",
  ".doc",
  ".exp",
  ".eps",
  ".h",
  ".hh",
  ".hpp",
  ".hxx",
  ".htm",
  ".html",
  ".inc",
  ".inf",
  ".ini",
  ".java",
  ".js",
  ".lnk",
  ".ltx",
  ".log",
  ".ly",
  ".mak",
  ".nas",
  ".pas",
  ".pl",
  ".pm",
  ".ps",
  ".pod",
  ".sh",
  ".sql",
  ".tex",
  ".tsv",
  ".txt",
  ".was",
  ".xsd",
  ".xsl",
  ".xslt",
  ".xml",
  "makefile",
  "readme",
  "todo",
  "changelog",
  "news",
};

/// number of strings in suffix array
const int suffixNum=sizeof(suffix)/sizeof(suffix[0]);

/** this value gives the relation from 'text characters' (<127) to
    'binary characters' (>127). The default value of .3 means that the file
    may include up to 30% of 'binary characters' before it is recognized as
    a binary file. If you do not want to check for binary files you can
    #undef the option DETECT_BINARY
*/
const float binaryDetect=.3;

/***************************************************************************
 *                                                                         *
 *            NO NEED TO CONFIGURE BEYOND THIS POINT                       *
 *                                                                         *
 ***************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

#include <functional>
#include <iostream>
#include <string>

using namespace std;

#define GETOPTEND EOF
const string DIRSEP="/";

unsigned char *in=0, *out=0;
size_t inSize=0, outSize=0;

/** converts the file 'filename' to unix-style textmode
    if the file was modified the original file is renamed with the suffix '.cr'<br>

    @param filename name of file
    @param checkBinary true if a binary check should be performed
    @return 0 on succes, -1 on error (mostly file errors)
*/
int killcr(const string& filename, const bool checkBinary)
{
  cerr << "killcr: checking " << filename << endl;

  // open the file
  FILE *f=fopen(filename.c_str(),"rb");
  if(f==0)
    {
      cerr << "killcr: error opening " << filename << endl;
      return -1;
    }

  // seek at the end of the file
  if(fseek(f,0,SEEK_END)==-1)
    {
      fclose(f);
      cerr << "killcr: error while getting filesize of " << filename << endl;
      return -1;
    }

  // now we get to know the filesize
  const size_t filesize=ftell(f);
  if(filesize==0)
    {
      fclose(f);
      cerr << "killcr: file size of 0: " << filename << endl;
      return -1;
    }

#if 0
  // allocate memory to read the file
  unsigned char *in=new unsigned char[filesize];
#else
  if(filesize>inSize)
    {
      inSize=filesize;
      in=reinterpret_cast<unsigned char*>(realloc(in, inSize));
    }
#endif
  if(in==0)
    {
      fclose(f);
      cerr << "killcr: could not allocate memory to read " << filename << endl;
      return -1;
    }

  // seek to the beginning of the file again
  if(fseek(f,0,SEEK_SET)==-1)
    {
      fclose(f);
      cerr << "killcr: could not seek to begin of " << filename << endl;
      delete [] in;
      return -1;
    }

  // read the complete file
  if(fread(in, 1, filesize, f)!=filesize)
    {
      fclose(f);
      cerr << "killcr: could not read " << filename << endl;
      delete [] in;
      return -1;
    }
  if(fclose(f))
    cerr << "killcr: could not close " << filename << endl;

  // allocate memory for the output text
#if 0
  unsigned char *out=new unsigned char[filesize];
#else
  if(filesize>outSize)
    {
      outSize=filesize;
      out=reinterpret_cast<unsigned char*>(realloc(out, outSize));
    }
#endif
  if(out==0)
    {
      cerr << "killcr: could not allocate memory: " << filesize << endl;
      delete [] in;
      return -1;
    }

  int outsize=filesize;  // size of the output text
  bool flag=false;       // flag if the output differs from the input
  int upper=0;           // count the number of characters below/above 127
  // process the input text
  for(size_t inpos=0, outpos=0 ; inpos<filesize; inpos++)
    {
      // if we encounter a <CR> which is followed by a <LF>
      if(in[inpos]==0xD && in[inpos+1]==0xA)
	{
	  // we make no output here, because we want to skip the <CR>
	  outsize--; // the output text does not contain the <CR> anymore
	  flag=true;    // we modified the text
	}
      else
	{
	  if(in[inpos]==0xD)
	    cerr << "<CR> detected, but not modified ";
	  out[outpos++]=in[inpos]; // copy the character
	}

      if(checkBinary)
	{
	  if(in[inpos]>127)          // count number of bytes above 127
	    upper++;

	  // after we read the first 256 bytes, we check for binary files
	  if(inpos>256)
	    if((float)upper/(float)inpos>binaryDetect)
	      {
		fprintf(stderr,"seems to be a binary file. Aborting!");
		flag=false;            // don't save the new file
		break;                 // exit the loop
	      }
	}
    }
#if 0
  delete [] in;  // dont need the input anymore
#endif
  // if the output text differs from the input text, we have to save it
  if(flag)
    {
      cerr << "gotcha you <CR>!";

      // we rename the old file
      string name=filename;
      size_t i=name.find_last_of('.');
      if(i!=string::npos && i>0)
	name.erase(i);
      name+=".cr";

      rename(filename.c_str(), name.c_str());

      // open the file again, this time writing
      f=fopen(filename.c_str(), "wb");
      if(f!=0)
	{
	  int resWrite=fwrite(out, outsize, 1, f);
	  if(resWrite!=1)
	    cerr << "killcr: could not write " << filename << endl;
	  fclose(f);
	}

    }
#if 0
  delete [] out;
#endif
  cerr << endl;
  return 0;
}

/** checks if filename contains a valid suffix. This
    function is not case sensitive!
    @param filename name of file
    @return true if 'filename' matches, false if 'filename' does not match
*/
bool check(const string& filename)
{
  // lowercase filename
  string s=filename;
  for(string::iterator i=s.begin(); i!=s.end(); ++i)
    *i=tolower(*i);

  // search for all suffixes
  for(int i=0; i<suffixNum; ++i)
    {
      const size_t p=s.rfind(suffix[i]);
      if(p==string::npos)
	continue;
      if(p == filename.size()-strlen(suffix[i]))
	return true;
    }

  return false;
}

/// file types
typedef enum {
  ERROR=-1,
  UNKNOWN=0,
  ISFILE,
  ISDIR,
} file_t;

/** returns the type of filename
    @param filename name of file
    @return ISFILE, ISDIR, UNKNOWN, ERROR
*/
file_t fileOrDir(const string& filename)
{
  struct stat filestat; // structure for a file status
  // if the status can not be fetched we skip processing of this file
  if(lstat(filename.c_str(), &filestat) == -1)
    return ERROR;
  // test for a normal file
  else if(S_ISREG(filestat.st_mode))
    return ISFILE;
  // test for a directory
  else if(S_ISDIR(filestat.st_mode))
    return ISDIR;

  return UNKNOWN;
}

/** checks if filename contains a valid suffix, then calls killcr(..)
    @param filename name of file
    @param checkSuffix true if suffix of filename should be checked against suffix list
    @param checkBinary true if file should be checked for binary data
    @return 0 if the file did not pass the 'int check(..)' test, otherwise
    the return value of 'int killcr(..)'
*/
int operate( const string& filename,
	     const bool checkSuffix,
	     const bool checkBinary
	     )
{
  if(checkSuffix)
    if(!check(filename))
      return 0;

  return killcr(filename, checkBinary);
}

/** parse through a directory and recursively into all
    subdirectories. For each file in the directory call the routine
    operate(char *filename)

    @param filename name of file
    @param checkSuffix true if suffix of filename should be checked against suffix list
    @param checkBinary true if file should be checked for binary data

    @return 0 if succesfully openend the directory, -1 if the directory could not be read/opened
*/
int parse( const string& filename,
	   const bool checkSuffix,
	   const bool checkBinary
	   )
{
  switch(fileOrDir(filename))
    {
    case ISFILE: operate(filename, checkSuffix, checkBinary); break;

    case ISDIR: DIR *dir; // structure to store a directory
      // open the directory
      if((dir=opendir(filename.c_str())) == 0)
	{
	  cerr << "parse: could not open dir " << filename << endl;
	  break;
	}

      // parse all entries in the directory
      struct dirent *file; // structure for a directory entry
      while((file=readdir(dir)) != 0)
	{

	  // the files '.' and '..' only cause trouble. We skip them
	  if( (strcmp(file->d_name, ".") == 0) ||
	      (strcmp(file->d_name, "..") == 0) )
	    continue;

	  parse(filename+DIRSEP+file->d_name, checkSuffix, checkBinary);
	}

      closedir(dir);
      break;
    default: break;
    }

  return 0;
}

/// prints the help screen to stdout
void help()
{
  printf("killcr:  a program to remove <CR> from text files\n");
  printf("usage:   KILLCR [options] [filenames]\n");
  printf("options: -h  show this text\n");
  printf("         -b  disable check for binary files\n");
  printf("         -c  disable file suffix check\n");
  printf("         -s  show valid file suffixes\n");
  printf("  without arguments the current directory is searched recursively for\n");
  printf("  files matching the valid suffixes.\n");
}

/// prints all suffixes to stdout
void suffixShow()
{
  for(int i=0; i<suffixNum; i++)
    cout << suffix[i] << endl;
}

/** parse through all directories via the 'int parse(..)' function. If no
    arguments are given use the current working directory. Otherwise the
    command line values are used.

    @return 0 when using the command line values;
    0 or -1 when using the working directory, depending on the parse(..) function
*/
int main(const int argc, char ** argv)
{
  bool checkSuffix=true;
  bool checkBinary=true;

  char opt=0;
  while(opt!=GETOPTEND)
    {
      opt=getopt(argc, argv, "cshb");
      switch(opt)
        {
	case 'h':
	case '?': help(); return 0;
	case 's': suffixShow(); return 0;
	case 'b': checkBinary=false; break;
	case 'c': checkSuffix=false; break;
        }
    }

  // if no command line arguments are given use the current directory
  if(optind==argc)
    {
      return parse(".", checkSuffix, checkBinary);
    }
  else
    // parse all command line arguments
    {
      while(optind<argc)
        {
          parse(argv[optind++], checkSuffix, checkBinary);
        }
    }

  if(in)
    free(in);
  if(out)
    free(out);

  return 0;
}
